/**
 ****************************************************************************************
 *
 * @file led_switch.c
 *
 * @brief LED Switch Service Server Implementation.
 *
 * Copyright (C) RivieraWaves 2009-2016
 *
 *
 ****************************************************************************************
 */

/**
 ****************************************************************************************
 * @addtogroup LED Switch
 * @{
 ****************************************************************************************
 */

/*
 * INCLUDE FILES
 ****************************************************************************************
 */
#include "rwip_config.h"

#if (BLE_LED_SWITCH_SERVER)
#include "attm.h"
#include "led_switch.h"
#include "led_switch_task.h"
#include "prf_utils.h"
#include "prf.h"
#include "ke_mem.h"
#include "attm_util128.h"

/*
 * MACROS
 ****************************************************************************************
 */

/// Maximal length for Characteristic values - 128 bytes
#define LED_SWITCH_VAL_MAX_LEN (128)

/*
 * CUSTOM ATTRIBUTES
 ****************************************************************************************
 */
static uint8_t led_svc_full_uuid[16] = LED_ATT_BT_UUID_128(ATT_LED_SWITCH_SVC_UUID);

/// Full LED Switch Database Description - Used to add attributes into the database
const struct attm_desc_128 led_switch_att_db[LED_SWTICH_IDX_NB] =
    {
        // Device Information Service Declaration
        [LED_SWITCH_IDX_SVC] = {ATT_BT_UUID_128(ATT_DECL_PRIMARY_SERVICE), PERM(RD, ENABLE), 0, 0},
        // Manufacturer Name Characteristic Declaration
        [LED_SWTICH_IDX_CHAR] = {ATT_BT_UUID_128(ATT_DECL_CHARACTERISTIC), PERM(RD, ENABLE), 0, 0},
        // Manufacturer Name Characteristic Value
        [LED_SWTICH_IDX_VAL] = {LED_ATT_BT_UUID_128(ATT_LED_SWITCH_CHAR_UUID), PERM(RD, ENABLE) | PERM(WRITE_COMMAND, ENABLE),
                                PERM(UUID_LEN, UUID_128) | PERM(RI, ENABLE) | ATT_UUID_128_LEN, LED_SWITCH_VAL_MAX_LEN},
};

/*
 * LOCAL FUNCTION DEFINITIONS
 ****************************************************************************************
 */

/**
 ****************************************************************************************
 * @brief Initialization of the LED switch module.
 * This function performs all the initializations of the Profile module.
 *  - Creation of database (if it's a service)
 *  - Allocation of profile required memory
 *  - Initialization of task descriptor to register application
 *      - Task State array
 *      - Number of tasks
 *      - Default task handler
 *
 * @param[out]    env        Collector or Service allocated environment data.
 * @param[in|out] start_hdl  Service start handle (0 - dynamically allocated), only applies for services.
 * @param[in]     app_task   Application task number.
 * @param[in]     sec_lvl    Security level (AUTH, EKS and MI field of @see enum attm_value_perm_mask)
 * @param[in]     param      Configuration parameters of profile collector or service (32 bits aligned)
 *
 * @return status code to know if profile initialization succeed or not.
 ****************************************************************************************
 */
static uint8_t led_switch_init(struct prf_task_env *env, uint16_t *start_hdl, uint16_t app_task, uint8_t sec_lvl, struct led_switch_db_cfg *params)
{
    //------------------ create the attribute database for the profile -------------------

    // Service content flag
    uint32_t cfg_flag;
    // DB Creation Statis
    uint8_t status = ATT_ERR_NO_ERROR;

    // Compute Attribute Table and save it in environment
    cfg_flag = led_switch_compute_cfg_flag(params->features);

    status = attm_util_svc_create_db128(start_hdl, led_svc_full_uuid, (uint8_t *)&cfg_flag,
                                        LED_SWTICH_IDX_NB, NULL, env->task, &led_switch_att_db[0],
                                        (sec_lvl & (PERM_MASK_SVC_DIS | PERM_MASK_SVC_AUTH | PERM_MASK_SVC_EKS)));

    //-------------------- allocate memory required for the profile  ---------------------
    if (status == ATT_ERR_NO_ERROR)
    {
        struct led_switch_env_tag *led_switch_env =
            (struct led_switch_env_tag *)ke_malloc(sizeof(struct led_switch_env_tag), KE_MEM_ATT_DB);

        // Remote I/O Level characteristic value permissions
        // uint16_t perm = PERM(RD, ENABLE) | PERM(WRITE_REQ, ENABLE);
        // attm_att_set_permission(*start_hdl + CUSTOM_IDX_REMOTE_IO_VAL, perm, 0);

        // allocate CUSTOM required environment variable
        env->env = (prf_env_t *)led_switch_env;
        led_switch_env->start_hdl = *start_hdl;
        led_switch_env->features = params->features;
        led_switch_env->prf_env.app_task = app_task | (PERM_GET(sec_lvl, SVC_MI) ? PERM(PRF_MI, ENABLE) : PERM(PRF_MI, DISABLE));
        led_switch_env->prf_env.prf_task = env->task | PERM(PRF_MI, DISABLE);

        // initialize environment variable
        env->id = TASK_ID_LED_SWITCH;
        env->desc.idx_max = LED_SWITCH_IDX_MAX;
        env->desc.state = led_switch_env->state;
        env->desc.default_handler = &led_switch_default_handler;
        co_list_init(&(led_switch_env->values));

        // service is ready, go into an Idle state
        ke_state_set(env->task, LED_SWITCH_IDLE);
    }

    return (status);
}
/**
 ****************************************************************************************
 * @brief Destruction of the CUSTOM module - due to a reset for instance.
 * This function clean-up allocated memory (attribute database is destroyed by another
 * procedure)
 *
 * @param[in|out]    env        Collector or Service allocated environment data.
 ****************************************************************************************
 */
static void led_switch_destroy(struct prf_task_env *env)
{
    struct led_switch_env_tag *led_switch_env = (struct led_switch_env_tag *)env->env;

    // remove all values present in list
    while (!co_list_is_empty(&(led_switch_env->values)))
    {
        struct co_list_hdr *hdr = co_list_pop_front(&(led_switch_env->values));
        ke_free(hdr);
    }

    // free profile environment variables
    env->env = NULL;
    ke_free(led_switch_env);
}

/**
 ****************************************************************************************
 * @brief Handles Connection creation
 *
 * @param[in|out]    env        Collector or Service allocated environment data.
 * @param[in]        conidx     Connection index
 ****************************************************************************************
 */
static void led_switch_create(struct prf_task_env *env, uint8_t conidx)
{
    /* Nothing to do */
}

/**
 ****************************************************************************************
 * @brief Handles Disconnection
 *
 * @param[in|out]    env        Collector or Service allocated environment data.
 * @param[in]        conidx     Connection index
 * @param[in]        reason     Detach reason
 ****************************************************************************************
 */
static void led_switch_cleanup(struct prf_task_env *env, uint8_t conidx, uint8_t reason)
{
    /* Nothing to do */
}

/*
 * GLOBAL VARIABLE DEFINITIONS
 ****************************************************************************************
 */

/// CUSTOM Task interface required by profile manager
const struct prf_task_cbs led_swtich_itf =
    {
        (prf_init_fnct)led_switch_init,
        led_switch_destroy,
        led_switch_create,
        led_switch_cleanup,
};

/*
 * FUNCTION DEFINITIONS
 ****************************************************************************************
 */

const struct prf_task_cbs *led_switch_prf_itf_get(void)
{
    return &led_swtich_itf;
}

uint32_t led_switch_compute_cfg_flag(uint16_t features)
{
    // Service Declaration
    uint32_t cfg_flag = 1;

    for (uint8_t i = 0; i < LED_SWITCH_CHAR_MAX; i++)
    {
        if (((features >> i) & 1) == 1)
        {
            cfg_flag |= (3 << (i * 2 + 1));
        }
    }

    return cfg_flag;
}

uint8_t led_switch_handle_to_value(struct led_switch_env_tag *env, uint16_t handle)
{
    uint8_t value = LED_SWITCH_CHAR_MAX;

    // handle cursor, start from first characteristic of service handle
    uint16_t cur_hdl = env->start_hdl + 1;

    for (uint8_t i = 0; i < LED_SWITCH_CHAR_MAX; i++)
    {
        if (((env->features >> i) & 1) == 1)
        {
            // check if value handle correspond to requested handle
            if ((cur_hdl + 1) == handle)
            {
                value = i;
                break;
            }
            cur_hdl += 2;
        }
    }

    return value;
}

uint16_t led_switch_value_to_handle(struct led_switch_env_tag *env, uint8_t value)
{
    uint16_t handle = env->start_hdl + 1;
    int8_t i;

    for (i = 0; i < LED_SWITCH_CHAR_MAX; i++)
    {
        if (((env->features >> i) & 1) == 1)
        {
            // requested value
            if (value == i)
            {
                handle += 1;
                break;
            }
            handle += 2;
        }
    }

    // check if handle found
    return ((i == LED_SWITCH_CHAR_MAX) ? ATT_INVALID_HDL : handle);
}

uint8_t led_switch_check_val_len(uint8_t char_code, uint8_t val_len)
{
    uint8_t status = GAP_ERR_NO_ERROR;

    // Check if length is upper than the general maximal length
    if (val_len > LED_SWITCH_VAL_MAX_LEN)
    {
        status = PRF_ERR_UNEXPECTED_LEN;
    }

    return (status);
}

#endif // BLE_LED_SWITCH_SERVER

/// @} LED_SWITCH
