/**
 ****************************************************************************************
 *
 * @file f000.c
 *
 * @brief f000 Server Implementation.
 *
 * Copyright (C) beken 2009-2015
 *
 *
 ****************************************************************************************
 */

#include "rwip_config.h"

#if (BLE_UART_TP_SERVER)
#include "attm.h"
#include "tp_mode.h"
#include "tp_mode_task.h"
#include "prf_utils.h"
#include "prf.h"
#include "ke_mem.h"
#include "attm_util128.h"
#include "uart.h"
#include "uart2.h"

/*
 * UART transparent ATTRIBUTES DEFINITION
 ****************************************************************************************
 */
static uint8_t tp_svc_full_uuid[16] = TP_ATT_BT_UUID_128(ATT_UART_TP_SERVER_SERVICE);
/// Full UART Tansparent Database Description - Used to add attributes into the database
const struct attm_desc_128 uart_tp_att_db[UART_TP_IDX_NB] =

    {
        // UART Transparent Service Declaration
        [UART_TP_IDX_SVC] = {ATT_BT_UUID_128(ATT_DECL_PRIMARY_SERVICE), PERM(RD, ENABLE), 0, 0},
        //  Characteristic Value
        [UART_TP_RX_CHAR] = {ATT_BT_UUID_128(ATT_DECL_CHARACTERISTIC), PERM(RD, ENABLE), 0, 0},
        // UART Transparent Rx Characteristic Value
        [UART_TP_RX_PAYLOAD] = {TP_ATT_BT_UUID_128(ATT_UART_TP_SERVER_CHAR_RX), PERM(WRITE_COMMAND, ENABLE), \
                                PERM(UUID_LEN, UUID_128) | PERM(RI, ENABLE) | ATT_UUID_128_LEN, UART_RX_PAYLOAD_LEN * sizeof(uint8_t)},
        //  Characteristic Value
        [UART_TP_TX_CHAR] = {ATT_BT_UUID_128(ATT_DECL_CHARACTERISTIC), PERM(RD, ENABLE), 0, 0},
        // UART Transparent Tx Characteristic Value
        [UART_TP_TX_PAYLOAD] = {TP_ATT_BT_UUID_128(ATT_UART_TP_SERVER_CHAR_TX), PERM(NTF, ENABLE), \
                                PERM(UUID_LEN, UUID_128) | PERM(RI, ENABLE) | ATT_UUID_128_LEN, UART_TX_PAYLOAD_LEN * sizeof(uint8_t)},
        // UART Transparent Characteristic - Client Characteristic Configuration Descriptor
        [UART_TP_TX_NTF_CFG] = {ATT_BT_UUID_128(ATT_DESC_CLIENT_CHAR_CFG), PERM(RD, ENABLE) | PERM(WRITE_REQ, ENABLE), 0, 0},

}; /// Macro used to retrieve permission value from access and rights on attribute.

static uint8_t uart_tp_init(struct prf_task_env *env, uint16_t *start_hdl, uint16_t app_task, uint8_t sec_lvl, struct uart_tp_db_cfg *params)
{
    uint16_t shdl;
    struct uart_tp_env_tag *uart_tp_env = NULL;
    // Status
    uint8_t status = GAP_ERR_NO_ERROR;

    //-------------------- allocate memory required for the profile  ---------------------
    uart_tp_env = (struct uart_tp_env_tag *)ke_malloc(sizeof(struct uart_tp_env_tag), KE_MEM_ATT_DB);
    memset(uart_tp_env, 0, sizeof(struct uart_tp_env_tag));

    // Service content flag
    uint8_t cfg_flag = UART_TP_CFG_FLAG_MANDATORY_MASK;

    // Save database configuration
    uart_tp_env->features |= (params->features);
    UART_PRINTF("%s,uart_tp_env->features : %x\r\n", __func__, uart_tp_env->features);
    // Check if notifications are supported
    if (params->features == UART_TP_TX_NTF_SUP)
    {
        cfg_flag |= UART_TP_CFG_FLAG_NTF_SUP_MASK;
    }
    shdl = *start_hdl;

    // Create UART transparent in the DB
    //------------------ create the attribute database for the profile -------------------
    status = attm_util_svc_create_db128(&(shdl), tp_svc_full_uuid, (uint8_t *)&cfg_flag,
                                UART_TP_IDX_NB, NULL, env->task, &uart_tp_att_db[0],
                                (sec_lvl & (PERM_MASK_SVC_DIS | PERM_MASK_SVC_AUTH | PERM_MASK_SVC_EKS)));

    //-------------------- Update profile task information  ---------------------
    if (status == ATT_ERR_NO_ERROR)
    {

        // allocate UART transparent required environment variable
        env->env = (prf_env_t *)uart_tp_env;
        *start_hdl = shdl;
        uart_tp_env->start_hdl = *start_hdl;
        uart_tp_env->prf_env.app_task = app_task | (PERM_GET(sec_lvl, SVC_MI) ? PERM(PRF_MI, ENABLE) : PERM(PRF_MI, DISABLE));
        uart_tp_env->prf_env.prf_task = env->task | PERM(PRF_MI, DISABLE);

        // initialize environment variable
        env->id = TASK_ID_UART_TP;
        env->desc.idx_max = UART_TP_IDX_MAX;
        env->desc.state = uart_tp_env->state;
        env->desc.default_handler = &uart_tp_default_handler;

        // service is ready, go into an Idle state
        ke_state_set(env->task, UART_TP_IDLE);
    }
    else if (uart_tp_env != NULL)
    {
        ke_free(uart_tp_env);
    }

    return (status);
}

static void uart_tp_destroy(struct prf_task_env *env)
{
    struct uart_tp_env_tag *uart_tp_env = (struct uart_tp_env_tag *)env->env;

    // clear on-going operation
    if (uart_tp_env->operation != NULL)
    {
        ke_free(uart_tp_env->operation);
    }

    // free profile environment variables
    env->env = NULL;
    ke_free(uart_tp_env);
}

static void uart_tp_create(struct prf_task_env *env, uint8_t conidx)
{
    struct uart_tp_env_tag *uart_tp_env = (struct uart_tp_env_tag *)env->env;
    ASSERT_ERR(conidx < BLE_CONNECTION_MAX);

    // force notification config to zero when peer device is connected
    uart_tp_env->ntf_cfg[conidx] = 0;
}

static void uart_tp_cleanup(struct prf_task_env *env, uint8_t conidx, uint8_t reason)
{
    struct uart_tp_env_tag *uart_tp_env = (struct uart_tp_env_tag *)env->env;

    ASSERT_ERR(conidx < BLE_CONNECTION_MAX);
    // force notification config to zero when peer device is disconnected
    uart_tp_env->ntf_cfg[conidx] = 0;
}

void uart_notify_tx_payload(struct uart_tp_env_tag *uart_tp_env, struct uart_tx_payload_upd_req const *param)
{
    // Allocate the GATT notification message
    struct gattc_send_evt_cmd *tx_payload = KE_MSG_ALLOC_DYN(GATTC_SEND_EVT_CMD,
                                                             KE_BUILD_ID(TASK_GATTC, 0), prf_src_task_get(&(uart_tp_env->prf_env), 0),
                                                             gattc_send_evt_cmd, sizeof(uint8_t) * (param->length));
    // Fill in the parameter structure
    tx_payload->operation = GATTC_NOTIFY;
    tx_payload->handle = uart_tp_get_att_handle(UART_TP_TX_PAYLOAD);
    // pack measured value in database
    tx_payload->length = param->length;
    // tx_payload->value[0] = uart_tp_env->tx_payload[0];
    memcpy(&tx_payload->value[0], &param->tx_payload[0], param->length);
    // send notification to peer device
    ke_msg_send(tx_payload);
}

/// BASS Task interface required by profile manager
const struct prf_task_cbs uart_tp_itf =
    {
        (prf_init_fnct)uart_tp_init,
        uart_tp_destroy,
        uart_tp_create,
        uart_tp_cleanup,
};

const struct prf_task_cbs *uart_tp_prf_itf_get(void)
{
    return &uart_tp_itf;
}

uint16_t uart_tp_get_att_handle(uint8_t att_idx)
{

    struct uart_tp_env_tag *uart_tp_env = PRF_ENV_GET(UART_TP, uart_tp);
    uint16_t handle = ATT_INVALID_HDL;

    handle = uart_tp_env->start_hdl;

    // increment index according to expected index
    if (att_idx < UART_TP_TX_NTF_CFG)
    {
        handle += att_idx;
    }
    // FFF1 notification
    else if ((att_idx == UART_TP_TX_NTF_CFG) && (((uart_tp_env->features) & 0x01) == UART_TP_TX_NTF_SUP))
    {
        handle += UART_TP_TX_NTF_CFG;
    }
    else
    {
        handle = ATT_INVALID_HDL;
    }

    return handle;
}

uint8_t uart_tp_get_att_idx(uint16_t handle, uint8_t *att_idx)
{
    struct uart_tp_env_tag *uart_tp_env = PRF_ENV_GET(UART_TP, uart_tp);
    uint16_t hdl_cursor = uart_tp_env->start_hdl;
    uint8_t status = PRF_APP_ERROR;

    // Browse list of services
    // handle must be greater than current index
    // check if it's a mandatory index
    if (handle <= (hdl_cursor + UART_TP_TX_PAYLOAD))
    {
        *att_idx = handle - hdl_cursor;
        status = GAP_ERR_NO_ERROR;
    }
    hdl_cursor += UART_TP_TX_PAYLOAD;

    // check if it's a notify index
    UART_PRINTF("features : %x \r\n", uart_tp_env->features);
    if (((uart_tp_env->features) & 0x01) == UART_TP_TX_NTF_SUP)
    {
        hdl_cursor++;
        if (handle == hdl_cursor)
        {
            *att_idx = UART_TP_TX_NTF_CFG;
            status = GAP_ERR_NO_ERROR;
        }
    }
    hdl_cursor++;

    return (status);
}

#endif // (BLE_UART_TP_SERVER)
