/**
  **************************************************************************
  * @file     main.c
  * @version  v2.0.3
  * @date     2022-08-16
  * @brief    main program
  **************************************************************************
  *                       Copyright notice & Disclaimer
  *
  * The software Board Support Package (BSP) that is made available to
  * download from Artery official website is the copyrighted work of Artery.
  * Artery authorizes customers to use, copy, and distribute the BSP
  * software and its related documentation for the purpose of design and
  * development in conjunction with Artery microcontrollers. Use of the
  * software is governed by this copyright notice and the following disclaimer.
  *
  * THIS SOFTWARE IS PROVIDED ON "AS IS" BASIS WITHOUT WARRANTIES,
  * GUARANTEES OR REPRESENTATIONS OF ANY KIND. ARTERY EXPRESSLY DISCLAIMS,
  * TO THE FULLEST EXTENT PERMITTED BY LAW, ALL EXPRESS, IMPLIED OR
  * STATUTORY OR OTHER WARRANTIES, GUARANTEES OR REPRESENTATIONS,
  * INCLUDING BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
  * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.
  *
  **************************************************************************
  */
  
/* Includes ------------------------------------------------------------------*/
#include "at32wb415_board.h"
#include "at32wb415_clock.h"
#include "stdio.h"
#include "string.h"
#include "at_cmd.h"
#include "tp_mode.h"
#include "usb_conf.h"
#include "usb_core.h"
#include "usbd_int.h"
#include "custom_hid_class.h"
#include "custom_hid_desc.h"

//#define BT_FLASH_WR_TEST
#define SIZEOFMSG   (sizeof(at_cmd_list)/sizeof(struct at_cmd_msg))

extern cmd_flash_read_status fr_flag;
extern cmd_rsp_status cmd_result;
extern flag_status recv_cmp_flag;
extern uint8_t bt_flash_data;
extern char recv_data[];
extern uartstruct tp_mode_rx_uart;
extern uartstruct tp_mode_tx_uart;

flag_status UART_TP_MODE = RESET;
flag_status USB_TP_MODE = RESET;

/* maybe this flag is not necessary */
flag_status switch_mode_flag = RESET;

uint32_t uart_tx_total_length = 0;

struct at_cmd_msg at_cmd_list[] =
{
  {AT_CMD_IO_SET_LOW, AT_CMD_IOSET0},
  {AT_CMD_IO_SET_HIGH, AT_CMD_IOSET1},
  {AT_CMD_IO_GET,  AT_CMD_IOGET},
};

static void uart_at_interface_init(uint32_t baudrate);
static void at_cmd_handler(void);
static void main_loop_func(void);

/* usb global struct define */
otg_core_type otg_core_struct;
uint8_t report_buf[USBD_CUSTOM_IN_MAXPACKET_SIZE];
void usb_clock48m_select(usb_clk48_s clk_s);
void usb_gpio_config(void);
void usb_low_power_wakeup_config(void);

/** @addtogroup AT32F415_StdPeriph_Examples
  * @{
  */

/** @addtogroup USART_Printf
  * @{
  */ 

/**
  * @brief  Main program
  * @param  None
  * @retval None
  */
int main(void)
{     
#ifdef BT_FLASH_WR_TEST
  uint8_t data;
#endif
  crm_clocks_freq_type crm_freq_struct = {0};
  system_clock_config();
  crm_clocks_freq_get(&crm_freq_struct);
  
  at32_board_init();
  at32_button_init();

  nvic_priority_group_config(NVIC_PRIORITY_GROUP_4);
  
  uart_at_interface_init(115200);
  
  uart_print_init(115200);
  printf("uart log ready...\r\n");
  
  /* Enable the USARTx Interrupt */
  usart_interrupt_enable(USART3, USART_RDBF_INT, TRUE);
  usart_enable(USART3, TRUE); 
  
  /* usb gpio config */
  usb_gpio_config();
  
  /* enable otgfs clock */
  crm_periph_clock_enable(OTG_CLOCK, TRUE);

  /* select usb 48m clcok source */
  usb_clock48m_select(USB_CLK_HEXT);

  /* enable otgfs irq */
  nvic_irq_enable(OTG_IRQ, 0, 0);

  /* init usb */
  usbd_init(&otg_core_struct,
            USB_FULL_SPEED_CORE_ID,
            USB_ID,
            &custom_hid_class_handler,
            &custom_hid_desc_handler);
            
  while (1)
  {
    main_loop_func();
  }
}

/**
  * @brief  this function config at-command interface
  * @param  baudrate: uart baudrate
  * @retval none
  */
static void uart_at_interface_init(uint32_t baudrate)
{
  gpio_init_type gpio_init_struct;
 
  nvic_irq_enable(USART3_IRQn, 0, 0);
  
  crm_periph_clock_enable(CRM_USART3_PERIPH_CLOCK, TRUE);
  crm_periph_clock_enable(CRM_GPIOA_PERIPH_CLOCK, TRUE);  
  crm_periph_clock_enable(CRM_IOMUX_PERIPH_CLOCK, TRUE);
  
  gpio_pin_remap_config(USART3_GMUX_0010, TRUE);
  gpio_init_struct.gpio_drive_strength = GPIO_DRIVE_STRENGTH_STRONGER;
  gpio_init_struct.gpio_out_type  = GPIO_OUTPUT_PUSH_PULL;
  gpio_init_struct.gpio_mode = GPIO_MODE_MUX;
  gpio_init_struct.gpio_pins = GPIO_PINS_7;
  gpio_init_struct.gpio_pull = GPIO_PULL_NONE;
  gpio_init(GPIOA, &gpio_init_struct);
  
  gpio_init_struct.gpio_drive_strength = GPIO_DRIVE_STRENGTH_STRONGER;
  gpio_init_struct.gpio_out_type  = GPIO_OUTPUT_PUSH_PULL;
  gpio_init_struct.gpio_mode = GPIO_MODE_INPUT;
  gpio_init_struct.gpio_pins = GPIO_PINS_6;
  gpio_init_struct.gpio_pull = GPIO_PULL_UP;
  gpio_init(GPIOA, &gpio_init_struct);
  
  usart_init(USART3, baudrate, USART_DATA_8BITS, USART_STOP_1_BIT);
  usart_transmitter_enable(USART3, TRUE);
  usart_receiver_enable(USART3, TRUE);
  
  /* Enable the USARTx Interrupt */
  usart_interrupt_enable(USART3, USART_RDBF_INT, TRUE);
  usart_enable(USART3, TRUE); 
}

/**
  * @brief  handle input at command, it will response corresponding formats.
  * @param  none
  * @retval none
  */
static void at_cmd_handler(void)
{
  uint8_t msg_id = SIZEOFMSG-1, i;
  if(recv_cmp_flag == SET)
  { 
      for(i = 0; i <= SIZEOFMSG; i++)
      {
        if(memcmp(recv_data, at_cmd_list[i].at_cmd_string, strlen(recv_data)) == 0)
        {
          //printf("%s", recv_data);
          msg_id = i;
          break;
        }
      }
      
      switch(at_cmd_list[msg_id].msg_id)
      {
        case AT_CMD_IOSET0:
        {
          at32_led_off(LED2);
          delay_us(250);
          at_cmd_send(AT_RESULT_OK0);
          break;
        }
        case AT_CMD_IOSET1:
        {
          at32_led_on(LED2);
          delay_us(250);
          at_cmd_send(AT_RESULT_OK1);
          break;
        }
        case AT_CMD_IOGET:
        {
          if(gpio_output_data_bit_read(GPIOB, GPIO_PINS_7))
          {
            at_cmd_send(AT_RESULT_OK1);
          }
          else
          {
            at_cmd_send(AT_RESULT_OK0);
          }
          break;
        }
        default:
        {
          at_cmd_send(AT_RSP_ERROR);
          break;
        }
      }          
    recv_cmp_flag = RESET;
    memset(recv_data, 0, strlen(recv_data));
  }
}

/**
  * @brief  this function handle general mode and transparent mode.
  * @param  none
  * @retval none
  */
static void main_loop_func(void)
{
  if(at32_button_press() == USER_BUTTON)
    {
      /* uart tranparent mode */
      if(UART_TP_MODE == RESET)
      {
        UART_TP_MODE = SET; 
      }
      /* usb transparent mode */
      else if(UART_TP_MODE == SET && USB_TP_MODE == RESET)
      {
        USB_TP_MODE = SET;
      }
      /* general mode */
      else
      {
        UART_TP_MODE = RESET;
        USB_TP_MODE = RESET;
      }
      
      if(UART_TP_MODE)
      {
        switch_mode_flag = SET;
        at_cmd_send(AT_CMD_TP_MODE_ON);
        while(cmd_rsp_get(AT_TP_MODE_OK1) != RSP_OK);
        switch_mode_flag = RESET;
        at32_led_on(LED3);
        if(USB_TP_MODE)
        {
          printf("enter usb tp mode\r\n");
        }
        else
        {
          printf("enter uart tp mode\r\n");
        }
      }
      else
      {        
        switch_mode_flag = SET;
        at_cmd_send(AT_CMD_TP_MODE_OFF);
        while(cmd_rsp_get(AT_TP_MODE_OK0) != RSP_OK);
        switch_mode_flag = RESET;
        at32_led_off(LED3);
        printf("enter general mode\r\n");
      }
    }    

    if(UART_TP_MODE)
    {
      tp_mode_rx_handler();
      tp_mode_tx_handler();
    }
    else
    {
      at_cmd_handler();
    }
}

/**
  * @brief  this function config gpio.
  * @param  none
  * @retval none
  */
void usb_gpio_config(void)
{
  gpio_init_type gpio_init_struct;

  crm_periph_clock_enable(OTG_PIN_GPIO_CLOCK, TRUE);
  gpio_default_para_init(&gpio_init_struct);

#ifdef USB_SOF_OUTPUT_ENABLE
  gpio_init_struct.gpio_drive_strength = GPIO_DRIVE_STRENGTH_STRONGER;
  gpio_init_struct.gpio_out_type  = GPIO_OUTPUT_PUSH_PULL;
  gpio_init_struct.gpio_mode = GPIO_MODE_MUX;
  gpio_init_struct.gpio_pull = GPIO_PULL_NONE;
  crm_periph_clock_enable(OTG_PIN_SOF_GPIO_CLOCK, TRUE);
  gpio_init_struct.gpio_pins = OTG_PIN_SOF;
  gpio_init(OTG_PIN_SOF_GPIO, &gpio_init_struct);
#endif
}

/**
  * @brief  usb 48M clock select
  * @param  clk_s:USB_CLK_HICK, USB_CLK_HEXT
  * @retval none
  */
void usb_clock48m_select(usb_clk48_s clk_s)
{
  switch(system_core_clock)
  {
    /* 48MHz */
    case 48000000:
      crm_usb_clock_div_set(CRM_USB_DIV_1);
      break;

    /* 72MHz */
    case 72000000:
      crm_usb_clock_div_set(CRM_USB_DIV_1_5);
      break;

    /* 96MHz */
    case 96000000:
      crm_usb_clock_div_set(CRM_USB_DIV_2);
      break;

    /* 120MHz */
    case 120000000:
      crm_usb_clock_div_set(CRM_USB_DIV_2_5);
      break;

    /* 144MHz */
    case 144000000:
      crm_usb_clock_div_set(CRM_USB_DIV_3);
      break;

    default:
      break;

  }
}

/**
  * @brief  this function handles otgfs interrupt.
  * @param  none
  * @retval none
  */
void OTG_IRQ_HANDLER(void)
{
  usbd_irq_handler(&otg_core_struct);
}

/**
  * @brief  usb delay millisecond function.
  * @param  ms: number of millisecond delay
  * @retval none
  */
void usb_delay_ms(uint32_t ms)
{
  /* user can define self delay function */
  delay_ms(ms);
}

/**
  * @brief  usb delay microsecond function.
  * @param  us: number of microsecond delay
  * @retval none
  */
void usb_delay_us(uint32_t us)
{
  delay_us(us);
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t* file, uint32_t line)
{ 
  /* User can add his own implementation to report the file name and line number,
     ex: printf("Wrong parameters value: file %s on line %d\r\n", file, line) */

  /* Infinite loop */
  while (1)
  {}
}

#endif

/**
  * @}
  */ 

/**
  * @}
  */ 
